//This is a web worker. It runs as a different "thread" from the main page UI thread.
//This web worker handles the requests for indications from the server (real QLCP target or developer tool) and sends controls.
//Received indications are parsed and each bit state is saved in the indicationBits array which is posted for the UI every other request interval.
//Each component in the UI is listening for the postmessage event from this web worker.

var requestInterval = 50;                   //this is the interval used for the periodic functions. It sets the base time in milliseconds for all periodicity
var flashTime = 10;                         //this is the flash interval which is multiplied times the requestInterval for determining total time
var flashTimeCntr = 0;                      //interval counter used to determine on/off flash duration
var reqIndFlag = true;                      //flag to determine whether to request indication or flash bits
var linkTimeout = 40;                       //link timeout = this value times the requestInterval
var linkTimeoutIntervalCntr = 0;            //interval counter to determine when to set link state to failed
var linkHealth = false;                     //link health state
var skipRequest = false;                    //this is for developer mode which does not request flash statuses
var indicationData = "";
var lastIndicationData = "";
var flashData = "";
var lastFlashData = "";

var requestActive = false;                  //don't send another request if this flag is true
var devMode = false;
var urlRoot = "";

var indicationBytes = new Array(64);        //array of indication bytes after indicationData is split
var indFlashBytes = new Array(64);          //array of indication flash bytes
var flashMaskBytes = new Uint8Array(64);
var flashOffState = false;
var indicationBits = new Array(512);
var ready = false;                          //true when ready to request indications from the server

//Add the event listener to process messages received from the UI thread.
//The message received from the main thread contains the following variables:
//   devMode - true when the web GUI is being used in the developer mode
//   url - IP address of the QLCP-NET entered in the browser
//   sendControl - true when the main thread is requesting an update control to be sent (one bit)
//   bitNbr - the bit number associated with the update control
//   bitState - true/false state of the bit associated with the update control
//   sendAllControls - true when the main thread is requesting a full control to be sent (all bits)
//   cntrlBits - array of the control bit states

this.addEventListener("message", function(e){

    if (e.data.devMode)
    {
        devMode = true;
        urlRoot = "http://localhost:9999/";
    }
    else
    {
        devMode = false;
        urlRoot = "http://" + e.data.url + ":54321/";
    }
        
    //Send a control if requested by the UI thread.   
    if (e.data.sendControl)
    {
        qnetSendControl(e.data.bitNbr, e.data.bitState);
    }
    
    if (e.data.sendAllControls)
    {
        qnetSendAllControls(e.data.cntrlBits);
    }
    
    console.log(e.data);
    
    //Now that we've received a message from the main thread defining the
    //developer mode and url, we are ready to request indications.
    ready = true;
        
});

//init flash bytes and flash mask
for (var i = 0; i < indFlashBytes.length; i++)
{
    indFlashBytes[i] = 0;  
    flashMaskBytes[i] = 0xFF;
}

RequestIndications();
                    
function RequestIndications()
{      
    var url = "";
    
    skipRequest = false;
    
    //if no indication data is received for link timeout time, show link failure icon in header
    if (linkTimeoutIntervalCntr >= linkTimeout)
    {
        linkHealth = false;

        //clear indication data
        indicationData = "";
        lastIndicationData = "";
        flashData = "";
        lastFlashData = "";
    }
    else
    {
        linkTimeoutIntervalCntr++;
    }
    
    if (ready)
    {
        //if there is not currently an active request...
        if (!requestActive)
        {
            //Assign url based on developer mode or time to request indication bits vs flash bits.
            //Every other request is for flashbits when connected to a QLCP-NET (not in developer mode).
            if (reqIndFlag)
            {
                reqIndFlag = false; //next time request flash bits
                url = urlRoot + "indications/";
            }
            else
            {
                reqIndFlag = true;  //next time request indication bits
                url = urlRoot + "flashbits/";
                
                //skip flash request if this is developer mode
                if (devMode) skipRequest = true;
            }

            if (!skipRequest)
            {
                requestActive = true;

                //Request the indication statuses 
                var Http = new XMLHttpRequest();
                Http.open("GET", url, true);
                Http.send();
                Http.onreadystatechange= function(){
                    //if request finished and response is ready...
                    if (this.readyState === 4)
                    {
                        if (this.status === 200) //OK
                        {
                            linkTimeoutIntervalCntr = 0;
                            linkHealth = true;
                            if (!reqIndFlag)
                            {
                                //save the message data
                                indicationData = Http.responseText; 

                                //Split the data into its separate bytes
                                indicationBytes = Http.responseText.split(",");
                            }
                            else
                            {
                                //save the message data
                                flashData = Http.responseText;

                                //Split the data into its separate bytes
                                indFlashBytes = Http.responseText.split(",");
                            }
                        }
                        requestActive = false;
                    }
                };
            }
        }

        if (flashTimeCntr === flashTime)
        {
            if (flashOffState === false)
            {
                flashOffState = true; 
                var i;
                var n = indFlashBytes.length;
                for (i=0; i < n; i++)
                {
                    //invert the bits
                    flashMaskBytes[i] = ~parseInt(indFlashBytes[i], 16);
                } 
            }
            else
            {
                var i;
                var n = indFlashBytes.length;
                flashOffState = false;
                for (i=0; i < n; i++)
                {
                    //invert the bits
                    flashMaskBytes[i] = 0xFF;
                } 
            } 
            
            flashTimeCntr = 0;
        }
        else
        {                    
            flashTimeCntr++;
        }
        
        //only process further and post a message if there has been a change since the last request.
        //Also, we only need to process after receiving both the data and flash data.
        //if (!reqIndFlag && ((indicationData !== lastIndicationData) || (flashData !== lastFlashData)))
        if (!reqIndFlag)
        {
            //apply flash mask to the indication bytes
            var i;
            var n = indicationBytes.length;
            for (i=0; i < n; i++)
            {
                indicationBytes[i] = (parseInt(indicationBytes[i],16) & flashMaskBytes[i]).toString(16);
            } 

            //console.log(indicationBytes);

            //unpack the bits
            for (i=0; i < indicationBits.length; i++)
            {
                indicationBits[i] = GetBitState(i);
            }

            //console.log(indicationBits);

            postMessage({indication:indicationBits, link:linkHealth});
                       
            lastIndicationData = indicationData;
            lastFlashData = flashData;
        }
    }    
    
    setTimeout("RequestIndications()", requestInterval);
    
}

//This function determines the bit state by parsing the indicationBytes array
function GetBitState(bit)
{               
    //start with LSB
    var bitmask = 1;

    //find the byte number
    var byteNbr = Math.floor(bit/8);

    //Find the remainder value and shift the bitmask that many times
    for (var i=0; i < bit % 8; i++)
    {
        //shift mask to left one bit
        bitmask = bitmask * 2;
    }

    //Convert the hex string to an integer and then AND it with
    //the bitmask to determine the bit state.
    if (parseInt(indicationBytes[byteNbr],16) & bitmask)
    {
        return 1; 
    }
    else
    {
        return 0;
    }                    
}

function qnetSendControl(bit, state)
{      
    var controlData = bit + "," + state;
    var Http = new XMLHttpRequest();
    Http.open("POST", urlRoot + "controls/bit/", true);
    Http.setRequestHeader("Content-type", "text/plain");
    Http.send(controlData);
    Http.onreadystatechange= function(){
        //not currently used
    };                           
}

function qnetSendAllControls(controlBits)
{
    //pack bits into comma separated control bytes
    var controlBytes = new Array(); //array to store the control bytes
    var controlData; //control string to send
    var byteValue = 0;

    //start with LSB
    var bitmask = 1; 
    var byteNbr = 0;

    //Find set bits and build the controlBytes array.
    for (var i=0; i < controlBits.length; i++)
    {                  
        if (controlBits[i])
        {
            byteValue = byteValue | bitmask;
        }

        //shift mask to left one bit
        bitmask = bitmask * 2;

        if (256 === bitmask)
        {
            //done with that byte, save it and prepare for next byte
            controlBytes.push(byteValue);
            bitmask = 1;
            byteNbr++;
            byteValue = 0;
        }     
    }

    //create control data string (convert to hex)
    controlData = controlBytes[0].toString(16);

    for (var i=1; i < controlBytes.length; i++)
    {
        controlData = controlData + "," + controlBytes[i].toString(16);
    }

    console.log(controlData);
    
    var Http = new XMLHttpRequest();
    Http.open("POST", urlRoot + "controls/all/", true);
    Http.setRequestHeader("Content-type", "text/plain");
    Http.send(controlData);
    Http.onreadystatechange= function(){
        //not currently used
    };
}